#include <VMA430_GPS.h>     // Include the GPS module library
#include <SoftwareSerial.h> // Include the software serial library
#include <Wire.h>
#include <SPI.h>
#include <SD.h>
#include <math.h>

#define PWM13 OCR4A

// SETUP I2C for the multiplexer
byte DRV = 0x5A; // DRV2605 slave address
byte ModeReg = 0x01;
byte TCA = 0x70;

SoftwareSerial ss(11, 12); // RX, TX
VMA430_GPS gps(&ss);     // Pass the softwareserial connection info the the GPS module library

// set up variables using the SD utility library functions:
Sd2Card card;
File sourceFile;
const int chipSelect = 10;
const int numCoordinates = 70;
float coordinates[numCoordinates];
int j =0;

int currentTargetIndex = 0;
float prevTargetLat = 0.0, prevTargetLon = 0.0;


int ports[2] = {1, 7}; // Adjust based on your connections
int motor1 = ports[0];
int motor2 = ports[1];

void setup()
{
  Wire.begin();
  Serial.begin(9600);
  while (!Serial) {
    ; // wait for serial port to connect. Needed for native USB port only
  }
  gps.begin(9600); // Sets up the GPS module to communicate with the Arduino over serial at 9600 baud
  gps.setUBXNav(); // Enable the UBX mavigation messages to be sent from the GPS module

  pwm13configure(); // sets up PWM signal
  delay(2);

  TCA9548A(motor1);
  initializeDRV2605();
  delay(10);
  pulse(0.1, 10);
  TCA9548A(motor2);
  initializeDRV2605();
  delay(10);
  pulse(0.1, 10);

  

  Serial.print("Initializing SD card...");

  if (!card.init(SPI_HALF_SPEED, chipSelect)) {
    Serial.println("initialization failed. Things to check:");
    Serial.println("* is a card inserted?");
    Serial.println("* is your wiring correct?");
    Serial.println("* did you change the chipSelect pin to match your shield or module?");
    while (1);
  } else {
    Serial.println("Wiring is correct and a card is present.");
  }

  if (!SD.begin(10)) {
    Serial.println("initialization failed!");
    while (1);
  }
  Serial.println("initialization done.");

  // open the file for reading:
  sourceFile = SD.open("test.txt");
  if (!sourceFile) {
    Serial.println("Failed to open source file for reading.");
    return;
  }

  if (sourceFile) {
    // read from the file until there's nothing else in it:
    // 
    while (sourceFile.available()) {
      String line = sourceFile.readStringUntil('\n');
      float numline = line.toFloat();
      coordinates[j] = numline;
      j++;
    }
  }
  
  // Close the file:
  sourceFile.close(); 
  // Print off the results
  //for(int i = 0; i < 50; i++) 
  //  Serial.println(sentences[i], 10);

  // Initialize previous target to the start target
  prevTargetLat = coordinates[0];
  prevTargetLon = coordinates[1];
}



void loop() {
  float targetLat, targetLon;
  if (readTargetCoordinates(&targetLat, &targetLon)) {
    float currentLat, currentLon;
    if (getGPSLocation(&currentLat, &currentLon)) {
      navigateToTarget(prevTargetLat, prevTargetLon, currentLat, currentLon, targetLat, targetLon);
      updateTargetIndex(); // Move to the next target
    }
  }
  delay(5000); // Update every 5 seconds
}

bool readTargetCoordinates(float* lat, float* lon) {
  if (currentTargetIndex < numCoordinates) {
    *lat = coordinates[currentTargetIndex];
    *lon = coordinates[currentTargetIndex + 1];
    return true;
  }
  return false;
}

bool getGPSLocation(float* latGPS, float* lonGPS) {
  if (gps.getUBX_packet()) // If a valid GPS UBX data packet is received...
  {
    gps.parse_ubx_data(); // Parse the new data
    *latGPS = gps.location.latitude;
    *lonGPS = gps.location.longitude;
    return true;
  }
  return false;
}


void updateTargetIndex() {
  // Move to the next target (even indices are latitudes, odd indices are longitudes)
  prevTargetLat = coordinates[currentTargetIndex];
  prevTargetLon = coordinates[currentTargetIndex + 1];
  currentTargetIndex += 2;
  if (currentTargetIndex >= numCoordinates) {
    currentTargetIndex = 0; // Loop back to the start if needed
  }
}

void navigateToTarget(float prevLat, float prevLon, float currentLat, float currentLon, float targetLat, float targetLon) {
  float threshold = 0.0001; // Threshold for reaching the target, this value equals around 10 meters

  // Check if current position is within the threshold of the target position
  if (abs(targetLat - currentLat) < threshold && abs(targetLon - currentLon) < threshold) {
    Serial.println("Reached the target position.");
    stop();
    return;
  }

  // Calculate the angle between the previous, current, and target positions
  float angle = calculateAngle(prevLat, prevLon, currentLat, currentLon, targetLat, targetLon);

  if (angle < 10) { // Nearly on the right path
    moveForward();
  } else if (angle > 170) { // Nearly opposite direction
    turnAround();
  } else {
    // Determine left or right based on the sign of the cross product
    if (isLeftTurn(prevLat, prevLon, currentLat, currentLon, targetLat, targetLon)) {
      moveLeft();
    } else {
      moveRight();
    }
  }
}

float calculateAngle(float prevLat, float prevLon, float currentLat, float currentLon, float nextLat, float nextLon) {
  // Vector A: previous target to current location
  float Ax = currentLon - prevLon;
  float Ay = currentLat - prevLat;
  
  // Vector B: current location to next target
  float Bx = nextLon - currentLon;
  float By = nextLat - currentLat;
  
  // Dot product of A and B
  float dotProduct = Ax * Bx + Ay * By;
  
  // Magnitudes of A and B
  float magnitudeA = sqrt(Ax * Ax + Ay * Ay);
  float magnitudeB = sqrt(Bx * Bx + By * By);
  
  // Calculate the cosine of the angle
  float cosTheta = dotProduct / (magnitudeA * magnitudeB);
  
  // Calculate the angle in radians and then convert to degrees
  float angle = acos(cosTheta);
  angle = degrees(angle);
  
  return angle;  // This is the angle between the two vectors
}

bool isLeftTurn(float prevLat, float prevLon, float currentLat, float currentLon, float nextLat, float nextLon) {
  // Vector A: previous target to current location
  float Ax = currentLon - prevLon;
  float Ay = currentLat - prevLat;
  
  // Vector B: current location to next target
  float Bx = nextLon - currentLon;
  float By = nextLat - currentLat;
  
  // Cross product of A and B (only the z-component)
  float crossProduct = Ax * By - Ay * Bx;
  
  // If the cross product is positive, the angle is to the left. If negative, it's to the right.
  return crossProduct > 0;
}

void moveForward() {
  Serial.println("Move forward");
  vibrateContinuous(motor1, 0.5, 5000); // Motor 1 vibrates continuously for 1 second
}

void moveRight() {
  Serial.println("Move right");
  vibratePattern(motor1, 0.7, 100, 50); // Motor 1 vibrates with a pattern
}

void moveLeft() {
  Serial.println("Move left");
  vibratePattern(motor2, 0.7, 100, 50); // Motor 2 vibrates with a pattern
}

void stop() {
  Serial.println("Stop!");
  vibrateContinuous(motor2, 0.5, 5000); // Motor 2 vibrates continuously for 5 seconds
}

void turnAround() {
  Serial.println("Turn around");
  vibrateAlternate(motor1, motor2, 1.0, 100, 50); // Motor 1 and Motor 2 alternate
}

void vibratePattern(int motor, double intensity, double duration, int dutyCycle) {
  TCA9548A(motor);
  vibrate(10, intensity, duration / 1000.0, dutyCycle); // Example vibration pattern
  vibrate(290, intensity, duration / 1000.0, dutyCycle);
  vibrate(10, intensity, duration / 1000.0, dutyCycle);
  vibrate(290, intensity, duration / 1000.0, dutyCycle);
}

void vibrateContinuous(int motor, double intensity, double duration) {
  TCA9548A(motor);
  long endTime = millis() + duration;
  while (millis() < endTime) {
    vibrate(290, intensity, duration / 1000.0, 100); // Continuous vibration
  }
}

void vibrateAlternate(int motor1, int motor2, double intensity, double duration, int dutyCycle) {
  long endTime = millis() + duration;
  while (millis() < endTime) {
    TCA9548A(motor1);
    vibrate(290, intensity, duration / 5000.0, dutyCycle);
    delay(duration / 5);
    TCA9548A(motor2);
    vibrate(290, intensity, duration / 5000.0, dutyCycle);
  }
}

//// Functions regarding the use of the motors, originated from the lab sessions
/*
void pulse(double intensity, double milliseconds)
pulse drives the hammer towards the closed end of the TacHammer
intensity defines the strength of the pulse. It ranges from [0-1] with 1 being strongest
When the hammer rebounds off of the repelling magnetic array, the inaudible pulse haptic
sensation is created milliseconds is the length of time the coil is activated in ms pulse is
intended to be sequenced with subsequent pulse and hit commands and if called on its own,
the hammer may travel after the rebound and strike the open end 
*/

void pulse(double intensity, double milliseconds) 
{
  int minimumint = 140;
  int maximumint = 255;
  int pwmintensity = (intensity * (maximumint - minimumint)) + minimumint;
  standbyOffB();
  PWM13 = pwmintensity;
  usdelay(milliseconds);
  standbyOnB();
}

/*
void pause(double milliseconds)
pause waits a duration of time defined in ms
*/
void pause(double milliseconds) 
{
  double us = milliseconds - ((int)milliseconds);
  standbyOnB();
  for (int i = 0; i <= milliseconds; i++) 
  {
    delay(1);
  }
  delayMicroseconds(us * 1000);
}

/*
void usdelay(double time)
usdelay implements a delay of the whole program in ms, it does not the same as pause
Can you describe the difference between the functions?
*/
void usdelay(double time) 
{
  double us = time - ((int)time);
  for (int i = 0; i <= time; i++) 
  {
    delay(1);
  }
  delayMicroseconds(us * 1000);
}

/*
void vibrate (double frequency, doub;le intensity, double duration, int dutycyle)
Vibrate repeatedly calls the pulse command to druve the TecHammer into the closed end of
the Techammer.
frequency defines the frequency of the vibration
intensity defines the strength of the vibrate. In ranges from 0 to 1, with 1 being the strongest.
duration defines the length of the vibration in s
dutycycle defines what percent of the perdio the TecHammer is active, it ranges from 0 to 100
Suggested duty cycles for vibrate
FREQ| 10| 30| 50| 70| 90|110|130|150|170|190|210|230|250|270|290
DUTY| 32| 28| 33| 36| 38| 37| 41| 47| 45| 47| 39| 40| 44| 43| 40
*/
void vibrate(double frequency, double intensity, double duration, int dutycycle) 
{
  int max_hit = 12;
  int min_hit = 1;
  int crossover = 60;
  int hitduration = 10 * dutycycle / frequency;

  boolean hold = false;
  double delayy;
  delayy = (1 / frequency * 1000) - hitduration;
  double timedown;
  timedown = duration * 1000;

  if (duration == 0)
  {
    hold =  true;
  }

  while (hold)
  {
    pulse(intensity, hitduration);
    pause(delayy);
  }

  while (timedown >= 0 && frequency < crossover)
  {
    pulse(intensity, hitduration);
    pause(3);
    pulse(0.002, delayy-3);

    timedown -= (delayy + hitduration);
  }

  while (timedown >= 0 && frequency >= crossover) 
  {
    pulse(intensity, hitduration);
    pause(delayy);
    timedown -= (delayy + hitduration);
  }
}

void standbyOnB() 
{
  Wire.beginTransmission(DRV);
  Wire.write(ModeReg);               // sets register pointer to the mode register (0x01)
  Wire.write(0x43);               // Puts the device pwm mode
  Wire.endTransmission();
}

void standbyOffB() 
{
  Wire.beginTransmission(DRV);
  Wire.write(ModeReg);               // sets register pointer to the mode register (0x01)
  Wire.write(0x03);               // Sets Waveform Mode to pwm
  Wire.endTransmission();
}

void initializeDRV2605() 
{
  Wire.beginTransmission(DRV);
  Wire.write(ModeReg);               // sets register pointer to the mode register (0x01)
  Wire.write(0x00);               // clear standby
  Wire.endTransmission();

  Wire.beginTransmission(DRV);
  Wire.write(0x1D);              // sets register pointer to the Libarary Selection register (0x1D)
  Wire.write(0xA8);              // set RTP unsigned
  Wire.endTransmission();

  Wire.beginTransmission(DRV);
  Wire.write(0x03);
  Wire.write(0x02);              // set to Library B, most aggresive
  Wire.endTransmission();

  Wire.beginTransmission(DRV);
  Wire.write(0x17);               // sets full scale reference
  Wire.write(0xff);               //
  Wire.endTransmission();

  Wire.beginTransmission(DRV);
  Wire.write(ModeReg);               // sets register pointer to the mode register (0x01)
  Wire.write(0x03);               // Sets Mode to pwm
  Wire.endTransmission();

  delay(100);

}

void TCA9548A(uint8_t bus) {
  Wire.beginTransmission(TCA);
  Wire.write(1 << bus);
  Wire.endTransmission();
}

// Configure the PWM clock
#define PWM12k  5   //  11719 Hz

void pwm13configure() {
  TCCR4A = 0;
  TCCR4B = PWM12k;
  TCCR4C = 0;
  TCCR4D = 0;
  PLLFRQ = (PLLFRQ & 0xCF) | 0x30;
  OCR4C = 255;
  pwmSet13();
}

// Set PWM to D13 (Timer4 A)
void pwmSet13() {
  OCR4A = 0;   
  DDRC |= _BV(7);    
  TCCR4A = 0x82;  
}